"""This module contains the FS URI resolver."""
from pathlib import Path

from polywrap_core import (
    FileReader,
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolver,
)
from polywrap_wasm import WasmPackage
from polywrap_wasm.constants import WRAP_MANIFEST_PATH, WRAP_MODULE_PATH


class SimpleFileReader(FileReader):
    """Defines a simple file reader."""

    def read_file(self, file_path: str) -> bytes:
        """Read a file.

        Args:
            file_path (str): The path of the file to read.

        Returns:
            bytes: The contents of the file.
        """
        with open(file_path, "rb") as f:
            return f.read()


class FsUriResolver(UriResolver):
    """Defines a URI resolver that resolves file system URIs.

    Args:
        file_reader (FileReader): The file reader used to read files.
    """

    file_reader: FileReader

    def __init__(self, file_reader: FileReader):
        """Initialize a new FsUriResolver instance."""
        self.file_reader = file_reader

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            UriPackageOrWrapper: The resolved URI.
        """
        if uri.authority not in ["fs", "file"]:
            return uri

        wrapper_path = Path(uri.path)

        wasm_module = self.file_reader.read_file(str(wrapper_path / WRAP_MODULE_PATH))

        manifest = self.file_reader.read_file(str(wrapper_path / WRAP_MANIFEST_PATH))

        return UriPackage(
            uri=uri,
            package=WasmPackage(
                wasm_module=wasm_module,
                manifest=manifest,
                file_reader=self.file_reader,
            ),
        )


__all__ = ["FsUriResolver", "SimpleFileReader"]



"""This module contains the UriResolverAggregator Resolver."""
# pylint: disable=unnecessary-ellipsis
from abc import ABC, abstractmethod
from typing import List, Optional

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
    UriWrapper,
)

from ...errors import UriResolutionError


class UriResolverAggregatorBase(UriResolver, ABC):
    """Defines a base resolver that aggregates a list of resolvers.
    
    This resolver aggregates a list of resolvers and tries to resolve\
        the uri with each of them. If a resolver returns a value\
        other than the resolving uri, the value is returned.
    """

    @abstractmethod
    def get_resolvers(
        self, client: InvokerClient, resolution_context: UriResolutionContext
    ) -> List[UriResolver]:
        """Get the list of resolvers to aggregate."""
        ...

    @abstractmethod
    def get_step_description(self) -> Optional[str]:
        """Get the description of the resolution step. Defaults to the class name."""
        ...

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        This method tries to resolve the uri with each of the aggregated\
            resolvers. If a resolver returns a value other than the resolving\
            uri, the value is returned.
        
        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for\
                resolving the URI.
            resolution_context (UriResolutionContext):\
                The resolution context to update.
        """
        sub_context = resolution_context.create_sub_history_context()

        for resolver in self.get_resolvers(client, sub_context):
            try:
                uri_package_or_wrapper = resolver.try_resolve_uri(
                    uri, client, sub_context
                )
            except UriResolutionError as e:
                step = UriResolutionStep(
                    source_uri=uri,
                    result=uri,
                    sub_history=sub_context.get_history(),
                    description=(
                        f"{self.get_step_description()} - Error: "
                        f"Failed to resolve uri: {uri}"
                    ),
                )
                resolution_context.track_step(step)
                raise e
            if (
                isinstance(uri_package_or_wrapper, (UriPackage, UriWrapper))
                or uri_package_or_wrapper != uri
            ):
                step = UriResolutionStep(
                    source_uri=uri,
                    result=uri_package_or_wrapper,
                    sub_history=sub_context.get_history(),
                    description=self.get_step_description(),
                )
                resolution_context.track_step(step)
                return uri_package_or_wrapper

        step = UriResolutionStep(
            source_uri=uri,
            result=uri,
            sub_history=sub_context.get_history(),
            description=self.get_step_description(),
        )
        resolution_context.track_step(step)
        return uri


__all__ = ["UriResolverAggregatorBase"]



"""This module contains the ResolverWithHistory abstract class."""
from abc import abstractmethod

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
)


class ResolverWithHistory(UriResolver):
    """Defines an abstract resolver that tracks its steps in\
        the resolution context.

    This is useful for resolvers that doesn't need to manually track \
        their steps in the resolution context.
    """

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI and \
            update the resolution context with the result.

        This method calls the internal abstract method _tryResolveUri before\
        updating the resolution context. Implementations are expect to place\
        resolution logic in _tryResolveUri.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for\
                resolving the URI.
            resolution_context (IUriResolutionContext):\
                The resolution context to update.

        Returns:
            UriPackageOrWrapper: The resolved URI package, wrapper, or URI.
        """
        result = self._try_resolve_uri(uri, client, resolution_context)
        step = UriResolutionStep(
            source_uri=uri, result=result, description=self.get_step_description()
        )
        resolution_context.track_step(step)

        return result

    @abstractmethod
    def get_step_description(self) -> str:
        """Get a description of the resolution step."""

    @abstractmethod
    def _try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Resolve a URI to a wrap package, a wrapper, or a URI using an internal function.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for\
                resolving the URI.
            resolution_context (IUriResolutionContext[UriPackageOrWrapper]):\
                The resolution context to update.
        """


__all__ = ["ResolverWithHistory"]



"""This module contains the legacy base URI resolver."""
from typing import Dict

from polywrap_core import (
    FileReader,
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolver,
)

from .fs_resolver import FsUriResolver
from .redirect_resolver import RedirectUriResolver


class BaseUriResolver(UriResolver):
    """Defines the base URI resolver.

    Args:
        file_reader (FileReader): The file reader to use.
        redirects (Dict[Uri, Uri]): The redirects to use.
    """

    _fs_resolver: FsUriResolver
    _redirect_resolver: RedirectUriResolver

    def __init__(self, file_reader: FileReader, redirects: Dict[Uri, Uri]):
        """Initialize a new BaseUriResolver instance."""
        self._fs_resolver = FsUriResolver(file_reader)
        self._redirect_resolver = RedirectUriResolver(redirects)

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            UriPackageOrWrapper: The resolved URI.
        """
        redirected_uri = self._redirect_resolver.try_resolve_uri(
            uri, client, resolution_context
        )

        return self._fs_resolver.try_resolve_uri(
            redirected_uri, client, resolution_context
        )


__all__ = ["BaseUriResolver"]



"""This module contains the type definition for StaticResolverLike.

StaticResolverLike is a type that represents a union of types\
    that can be used as a StaticResolver.
"""
from typing import Dict

from polywrap_core import Uri, UriPackageOrWrapper

StaticResolverLike = Dict[Uri, UriPackageOrWrapper]



"""This module contains the RedirectUriResolver class."""
from typing import Dict

from polywrap_core import InvokerClient, Uri, UriResolutionContext, UriResolver


class RedirectUriResolver(UriResolver):
    """Defines the redirect URI resolver.

    Args:
        redirects (Dict[Uri, Uri]): The redirects to use.
    """

    _redirects: Dict[Uri, Uri]

    def __init__(self, redirects: Dict[Uri, Uri]):
        """Initialize a new RedirectUriResolver instance."""
        self._redirects = redirects

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> Uri:
        """Try to resolve a URI to redirected URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            Uri: The resolved URI.
        """
        return self._redirects[uri] if uri in self._redirects else uri


__all__ = ["RedirectUriResolver"]



# pylint: disable=line-too-long
"""This package contains URI resolvers for polywrap-client.

Resolvers
---------
.. csv-table::
    :header: "resolver", "description"

    "WrapperResolver", "Defines a simple statically registered resolver for a wrapper."
    "PackageResolver", "Defines a simple statically registered resolver for a wrap package."
    "RedirectResolver", "Defines a simple resolver to redirect a URI to another URI."
    "StaticResolver", "Defines a simple resolver that allows registering an Uri to redirect or resolve to wrapper or wrap package."
    "UriResolverAggregator", "Defines a resolver that aggregates a list of resolvers."
    "RecursiveResolver", "Defines a resolver that recursively resolves the URI until the result is no longer a URI."
    "ExtendableUriResolver", "Defines a resolver that resolves a uri to a wrapper by using extension wrappers."
    "ResolutionResultCacheResolver", "Defines a resolver that caches the URI resolution result."

.. csv-table::
    :header: "error", "description"

    "UriResolutionError", "Base class for all errors related to URI resolution."
    "InfiniteLoopError", "Raised when an infinite loop is detected while resolving a URI."
    "UriResolverExtensionError", "Base class for all errors related to URI resolver extensions."
    "UriResolverExtensionNotFoundError", "Raised when an extension resolver wrapper could not be found for a URI."
"""
from .errors import *
from .resolvers import *
from .types import *



"""This module contains the RedirectResolver class."""
from polywrap_core import InvokerClient, Uri, UriPackageOrWrapper, UriResolutionContext

from ..abc import ResolverWithHistory


class RedirectResolver(ResolverWithHistory):
    """Defines a resolver that redirects a uri to another uri.

    This resolver redirects a uri to another uri. If the uri to resolve is the same as the\
        uri to redirect from, the uri to redirect to is returned. Otherwise, the uri to resolve\
        is returned.

    Args:
        from_uri (Uri): The uri to redirect from.
        to_uri (Uri): The uri to redirect to.
    """

    __slots__ = ("from_uri", "to_uri")

    from_uri: Uri
    to_uri: Uri

    def __init__(self, from_uri: Uri, to_uri: Uri) -> None:
        """Initialize a new RedirectResolver instance."""
        self.from_uri = from_uri
        self.to_uri = to_uri
        super().__init__()

    def get_step_description(self) -> str:
        """Get the description of the resolver step.

        Returns:
            str: The description of the resolver step.
        """
        return f"Redirect ({self.from_uri} - {self.to_uri})"

    def _try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        This method tries to resolve the given uri to a wrap package, a wrapper, or a URI.\
            If the given uri is the same as the uri to redirect from, the uri to redirect to is\
            returned. Otherwise, the given uri is returned.
        
        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.
        
        Returns:
            UriPackageOrWrapper: The resolved URI package, wrapper, or URI.
        """
        return uri if uri != self.from_uri else self.to_uri


__all__ = ["RedirectResolver"]



"""This module contains the WrapperCacheResolver."""
from dataclasses import dataclass
from typing import List, Optional, Union

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
    UriWrapper,
)
from polywrap_manifest import DeserializeManifestOptions

from .wrapper_cache import WrapperCache


@dataclass(kw_only=True, slots=True)
class WrapperCacheResolverOptions:
    """Defines the options for the WrapperCacheResolver.

    Args:
        deserialize_manifest_options (DeserializeManifestOptions): The options\
            to use when deserializing the manifest.
        end_on_redirect (Optional[bool]): Whether to end the resolution\
            process when a redirect is encountered. Defaults to False.
    """

    deserialize_manifest_options: DeserializeManifestOptions
    end_on_redirect: Optional[bool]


class WrapperCacheResolver(UriResolver):
    """Defines a resolver that caches wrappers by uri.

    This resolver caches the results of URI Resolution.\
        If result is an uri or package, it returns it back without caching.\
        If result is a wrapper, it caches the wrapper and returns it back.

    Args:
        resolver_to_cache (UriResolver): The URI resolver to cache.
        cache (WrapperCache): The cache to use.
        options (Optional[WrapperCacheResolverOptions]): The options to use.
    """

    __slots__ = ("resolver_to_cache", "cache", "options")

    resolver_to_cache: UriResolver
    cache: WrapperCache
    options: Optional[WrapperCacheResolverOptions]

    def __init__(
        self,
        resolver_to_cache: UriResolver,
        cache: WrapperCache,
        options: Optional[WrapperCacheResolverOptions] = None,
    ):
        """Initialize a new PackageToWrapperCacheResolver instance."""
        self.resolver_to_cache = resolver_to_cache
        self.cache = cache
        self.options = options

    def get_options(self) -> Union[WrapperCacheResolverOptions, None]:
        """Get the options of the resolver.

        Returns:
            CacheResolverOptions: The options of the resolver.
        """
        return self.options

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrapper, or a URI.

        This method tries to resolve the uri with the resolver to cache.\
            If the result is an uri or package, it returns it back without caching.\
            If the result is a wrapper, it caches the wrapper and returns it back.
        
        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use.
            resolution_context (UriResolutionContext): The resolution\
                context to use.
        
        Returns:
            UriPackageOrWrapper: The result of the resolution.
        """
        if uri_wrapper := self.cache.get(uri):
            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=uri_wrapper,
                    description="WrapperCacheResolver (Cache Hit)",
                )
            )
            return uri_wrapper

        sub_context = resolution_context.create_sub_history_context()

        result = self.resolver_to_cache.try_resolve_uri(
            uri,
            client,
            sub_context,
        )

        if isinstance(result, UriWrapper):
            resolution_path: List[Uri] = sub_context.get_resolution_path()

            for cache_uri in resolution_path:
                self.cache.set(cache_uri, result)

        resolution_context.track_step(
            UriResolutionStep(
                source_uri=uri,
                result=result,
                sub_history=sub_context.get_history(),
                description="WrapperCacheResolver (Cache Miss)",
            )
        )
        return result


__all__ = ["WrapperCacheResolver", "WrapperCacheResolverOptions"]



"""This module contains the PackageToWrapperResolver class."""
from dataclasses import dataclass
from typing import Optional

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
    UriWrapper,
)
from polywrap_manifest import DeserializeManifestOptions

from ..abc import ResolverWithHistory


@dataclass(kw_only=True, slots=True)
class PackageToWrapperResolverOptions:
    """Defines the options for the PackageToWrapperResolver.

    Args:
        deserialize_manifest_options (DeserializeManifestOptions): The options\
            to use when deserializing the manifest.
    """

    deserialize_manifest_options: Optional[DeserializeManifestOptions]


class PackageToWrapperResolver(ResolverWithHistory):
    """Defines a resolver that converts packages to wrappers.

    This resolver converts packages to wrappers.\
        If result is an uri, it returns it back.\
        If result is a wrapper, it returns it back.\
        In case of a package, it creates a wrapper and returns it back.
    
    Args:
        resolver (UriResolver): The URI resolver to cache.
        options (PackageToWrapperResolverOptions): The options to use.
    """

    resolver: UriResolver
    options: Optional[PackageToWrapperResolverOptions]

    def __init__(
        self,
        resolver: UriResolver,
        options: Optional[PackageToWrapperResolverOptions] = None,
    ) -> None:
        """Initialize a new PackageToWrapperResolver instance."""
        self.resolver = resolver
        self.options = options
        super().__init__()

    def _try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve the given URI to a wrapper or a redirected URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use.
            resolution_context (IUriResolutionContext):\
                The resolution context to use.
        
        Returns:
            UriPackageOrWrapper: The resolved URI or wrapper.
        """
        sub_context = resolution_context.create_sub_context()
        result = self.resolver.try_resolve_uri(uri, client, sub_context)
        if isinstance(result, UriPackage):
            wrapper = result.package.create_wrapper()
            result = UriWrapper(uri=uri, wrapper=wrapper)

        resolution_context.track_step(
            UriResolutionStep(
                source_uri=uri,
                result=result,
                sub_history=sub_context.get_history(),
                description=self.get_step_description(),
            )
        )
        return result

    def get_step_description(self) -> str:
        """Get the description of the resolution step.

        Returns:
            str: The description of the resolution step.
        """
        return self.__class__.__name__


__all__ = ["PackageToWrapperResolver", "PackageToWrapperResolverOptions"]



"""This package contains the resolvers for the Polywrap client."""
from .abc import *
from .aggregator import *
from .cache import *
from .extensions import *
from .legacy import *
from .package import *
from .recursive import *
from .redirect import *
from .static import *
from .wrapper import *



"""This module contains the in-memory wrapper cache."""
from typing import Dict, Union

from polywrap_core import Uri, UriPackageOrWrapper

from ....errors import UriResolutionError
from .resolution_result_cache import ResolutionResultCache


class InMemoryResolutionResultCache(ResolutionResultCache):
    """InMemoryResolutionResultCache is an in-memory implementation \
        of the resolution result cache protocol."""

    map: Dict[Uri, Union[UriPackageOrWrapper, UriResolutionError]]
    """The map of uris to resolution result."""

    def __init__(self):
        """Initialize a new InMemoryResolutionResultCache instance."""
        self.map = {}

    def get(self, uri: Uri) -> Union[UriPackageOrWrapper, UriResolutionError, None]:
        """Get the resolution result from the cache by its uri."""
        return self.map.get(uri)

    def set(
        self, uri: Uri, result: Union[UriPackageOrWrapper, UriResolutionError]
    ) -> None:
        """Set the resolution result in the cache by its uri."""
        self.map[uri] = result

    def __str__(self) -> str:
        """Display cache as a string."""
        return f"{self.map}"


__all__ = ["InMemoryResolutionResultCache"]



"""This module contains the UriResolverExtensionFileReader class."""
from pathlib import Path

from polywrap_core import FileReader, Invoker, Uri


class UriResolverExtensionFileReader(FileReader):
    """Defines a file reader that uses an extension wrapper to read files.

    This file reader uses an extension wrapper to read files.\
        The extension wrapper is used to read files by invoking the getFile method.\
        The getFile method is invoked with the path of the file to read.

    Args:
        extension_uri (Uri): The uri of the extension wrapper.
        wrapper_uri (Uri): The uri of the wrapper that uses the extension wrapper.
        invoker (Invoker): The invoker used to invoke the getFile method.
    """

    extension_uri: Uri
    """The uri of the extension wrapper."""

    wrapper_uri: Uri
    """The uri of the wrapper that uses the extension wrapper."""

    invoker: Invoker
    """The invoker used to invoke the getFile method."""

    def __init__(
        self,
        extension_uri: Uri,
        wrapper_uri: Uri,
        invoker: Invoker,
    ):
        """Initialize a new UriResolverExtensionFileReader instance."""
        self.extension_uri = extension_uri
        self.wrapper_uri = wrapper_uri
        self.invoker = invoker

    def read_file(self, file_path: str) -> bytes:
        """Read a file using the extension wrapper.

        Args:
            file_path (str): The path of the file to read.

        Returns:
            bytes: The contents of the file.
        """
        path = str(Path(self.wrapper_uri.path).joinpath(file_path))
        result = self.invoker.invoke(
            uri=self.extension_uri, method="getFile", args={"path": path}
        )

        if not isinstance(result, bytes):
            raise FileNotFoundError(
                f"File not found at path: {path}, using resolver: {self.extension_uri}"
            )
        return result


__all__ = ["UriResolverExtensionFileReader"]



"""This module contains the wrapper cache interface."""
from abc import abstractmethod
from typing import Protocol, Union

from polywrap_core import Uri, UriWrapper


class WrapperCache(Protocol):
    """Defines a cache interface for caching wrappers by uri.

    This is used by the wrapper resolver to cache wrappers for a given uri.
    """

    @abstractmethod
    def get(self, uri: Uri) -> Union[UriWrapper, None]:
        """Get a wrapper from the cache by its uri."""

    @abstractmethod
    def set(self, uri: Uri, wrapper: UriWrapper) -> None:
        """Set a wrapper in the cache by its uri."""


__all__ = ["WrapperCache"]



"""This module contains the UriResolverAggregator Resolver."""
from typing import List, Optional

from polywrap_core import InvokerClient, UriResolutionContext, UriResolver

from .uri_resolver_aggregator_base import UriResolverAggregatorBase


class UriResolverAggregator(UriResolverAggregatorBase):
    """Defines a resolver that aggregates a list of resolvers.
    
    This resolver aggregates a list of resolvers and tries to resolve\
        the uri with each of them. If a resolver returns a value\
        other than the resolving uri, the value is returned.

    Args:
        resolvers (List[UriResolver]): The list of resolvers to aggregate.
        step_description (Optional[str]): The description of the resolution\
            step. Defaults to the class name.
    """

    __slots__ = ("_resolvers", "_step_description")

    _resolvers: List[UriResolver]
    _step_description: Optional[str]

    def __init__(
        self, resolvers: List[UriResolver], step_description: Optional[str] = None
    ):
        """Initialize a new UriResolverAggregator instance."""
        self._step_description = step_description or self.__class__.__name__
        self._resolvers = resolvers
        super().__init__()

    def get_step_description(self) -> Optional[str]:
        """Get the description of the resolution step."""
        return self._step_description

    def get_resolvers(
        self, client: InvokerClient, resolution_context: UriResolutionContext
    ) -> List[UriResolver]:
        """Get the list of resolvers to aggregate."""
        return self._resolvers


__all__ = ["UriResolverAggregator"]



"""This module contains the recursive resolver."""
from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolver,
)

from ...errors import InfiniteLoopError


class RecursiveResolver(UriResolver):
    """Defines the recursive resolver.

    The recursive resolver is a wrapper around another resolver that\
    recursively resolves the URI until the result is no longer a URI.

    Args:
        resolver (UriResolver): The resolver to use.
    """

    __slots__ = ("resolver",)

    resolver: UriResolver

    def __init__(self, resolver: UriResolver):
        """Initialize a new RecursiveResolver instance."""
        self.resolver = resolver

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            UriPackageOrWrapper: The resolved URI.
        """
        if resolution_context.is_resolving(uri):
            raise InfiniteLoopError(uri)

        resolution_context.start_resolving(uri)

        uri_package_or_wrapper = self.resolver.try_resolve_uri(
            uri, client, resolution_context
        )

        if isinstance(uri_package_or_wrapper, Uri) and uri_package_or_wrapper != uri:
            uri_package_or_wrapper = self.try_resolve_uri(
                uri_package_or_wrapper,
                client,
                resolution_context,
            )

        resolution_context.stop_resolving(uri)

        return uri_package_or_wrapper


__all__ = ["RecursiveResolver"]



"""This module contains the PackageResolver class."""
from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    WrapPackage,
)

from ..abc import ResolverWithHistory


class PackageResolver(ResolverWithHistory):
    """Defines a resolver that resolves a uri to a package.

    Args:
        uri (Uri): The uri to resolve.
        package (WrapPackage): The wrap package to return.
    """

    __slots__ = ("uri", "package")

    uri: Uri
    package: WrapPackage

    def __init__(self, uri: Uri, package: WrapPackage):
        """Initialize a new PackageResolver instance."""
        self.uri = uri
        self.package = package
        super().__init__()

    def get_step_description(self) -> str:
        """Get the description of the resolver step.

        Returns:
            str: The description of the resolver step.
        """
        return f"Package ({self.uri.uri})"

    def _try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        This method tries to resolve the given uri to a wrap package, a wrapper, or a URI.\
            If the given uri is the same as the uri of the resolver, the wrap package is returned.\
            Otherwise, the given uri is returned.
        
        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for\
                resolving the URI.
            resolution_context (UriResolutionContext): The\
                resolution context.
        
        Returns:
            UriPackageOrWrapper: The resolved URI package, wrapper, or URI.
        """
        return uri if uri != self.uri else UriPackage(uri=uri, package=self.package)


__all__ = ["PackageResolver"]



"""This module contains the ExtendableUriResolver class."""
from typing import List, Optional

from polywrap_core import InvokerClient, Uri, UriResolutionContext, UriResolver

from ..aggregator import UriResolverAggregatorBase
from .extension_wrapper_uri_resolver import ExtensionWrapperUriResolver


class ExtendableUriResolver(UriResolverAggregatorBase):
    """Defines a resolver that resolves a uri to a wrapper by using extension wrappers.

    This resolver resolves a uri to a wrapper by using extension wrappers.\
        The extension wrappers are resolved using the extension wrapper uri resolver.\
        The extension wrappers are aggregated using the uri resolver aggregator.\
        The aggregated extension wrapper resolver is then used to resolve\
        the uri to a wrapper.

    Args:
        ext_interface_uris (Optional[List[Uri]]): The list of extension\
            interface uris. Defaults to the default list of extension\
            interface uris.
        resolver_name (Optional[str]): The name of the resolver. Defaults\
            to the class name.
    """

    DEFAULT_EXT_INTERFACE_URIS = [
        Uri.from_str("wrap://ens/wraps.eth:uri-resolver-ext@1.1.0"),
        Uri.from_str("wrap://ens/wraps.eth:uri-resolver-ext@1.0.0"),
        Uri.from_str("wrapscan.io/polywrap/uri-resolver@1.0"),
    ]
    """The default list of extension interface uris."""

    ext_interface_uris: List[Uri]
    """The list of extension interface uris."""

    resolver_name: str
    """The name of the resolver."""

    def __init__(
        self,
        ext_interface_uris: Optional[List[Uri]] = None,
        resolver_name: Optional[str] = None,
    ):
        """Initialize a new ExtendableUriResolver instance."""
        self.ext_interface_uris = ext_interface_uris or self.DEFAULT_EXT_INTERFACE_URIS
        self.resolver_name = resolver_name or self.__class__.__name__
        super().__init__()

    def get_step_description(self) -> Optional[str]:
        """Get the description of the resolution step."""
        return self.resolver_name

    def get_resolvers(
        self, client: InvokerClient, resolution_context: UriResolutionContext
    ) -> List[UriResolver]:
        """Get the list of resolvers to aggregate."""
        uri_resolvers_uris: List[Uri] = []

        for ext_interface_uri in self.ext_interface_uris:
            uri_resolvers_uris.extend(
                client.get_implementations(ext_interface_uri, apply_resolution=False)
                or []
            )

        return [
            ExtensionWrapperUriResolver(uri)
            for uri in uri_resolvers_uris
            if not resolution_context.is_resolving(uri)
        ]


__all__ = ["ExtendableUriResolver"]



"""This module contains the in-memory wrapper cache."""
from typing import Dict, Union

from polywrap_core import Uri, UriWrapper

from .wrapper_cache import WrapperCache


class InMemoryWrapperCache(WrapperCache):
    """InMemoryWrapperCache is an in-memory implementation\
        of the wrapper cache interface."""

    map: Dict[Uri, UriWrapper]
    """The map of uris to wrappers."""

    def __init__(self):
        """Initialize a new InMemoryWrapperCache instance."""
        self.map = {}

    def get(self, uri: Uri) -> Union[UriWrapper, None]:
        """Get a wrapper from the cache by its uri."""
        return self.map.get(uri)

    def set(self, uri: Uri, wrapper: UriWrapper) -> None:
        """Set a wrapper in the cache by its uri."""
        self.map[uri] = wrapper


__all__ = ["InMemoryWrapperCache"]



"""This module contains the ExtensionWrapperUriResolver class."""
from __future__ import annotations

from typing import Optional, TypedDict

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
    WrapError,
)
from polywrap_wasm import WasmPackage

from ...errors import (
    InfiniteLoopError,
    UriResolverExtensionError,
    UriResolverExtensionNotFoundError,
)
from .uri_resolver_extension_file_reader import UriResolverExtensionFileReader


class MaybeUriOrManifest(TypedDict, total=False):
    """Defines a type for the return value of the extension wrapper's\
        tryResolveUri function.

    The extension wrapper's tryResolveUri function can return either a uri\
        or a manifest. This type defines the return value of the function.
    """

    uri: Optional[str]
    manifest: Optional[bytes]


class ExtensionWrapperUriResolver(UriResolver):
    """Defines a resolver that resolves a uri to a wrapper by using an extension wrapper.

    This resolver resolves a uri to a wrapper by using an extension wrapper.\
        The extension wrapper is resolved using the extension wrapper uri resolver.\
        The extension wrapper is then used to resolve the uri to a wrapper.

    Args:
        extension_wrapper_uri (Uri): The uri of the extension wrapper.
    """

    __slots__ = ("extension_wrapper_uri",)

    extension_wrapper_uri: Uri
    """The uri of the extension wrapper."""

    def __init__(self, extension_wrapper_uri: Uri):
        """Initialize a new ExtensionWrapperUriResolver instance."""
        self.extension_wrapper_uri = extension_wrapper_uri

    def get_step_description(self) -> str:
        """Get the description of the resolver step.

        Returns:
            str: The description of the resolver step.
        """
        return f"ResolverExtension ({self.extension_wrapper_uri})"

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        This method tries to resolve the uri using the extension wrapper.\
            If the extension wrapper returns a uri, the uri is returned.\
            If the extension wrapper returns a manifest, the manifest is used\
            to create a wrapper and the wrapper is returned.
        
        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for\
                resolving the URI.
            resolution_context (UriResolutionContext): The\
                resolution context.
        
        Returns:
            UriPackageOrWrapper: The resolved URI package, wrapper, or URI.
        """
        sub_context = resolution_context.create_sub_context()

        try:
            uri_package_or_wrapper = self._try_resolve_uri_with_extension(
                uri, client, sub_context
            )

            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=uri_package_or_wrapper,
                    description=self.get_step_description(),
                    sub_history=sub_context.get_history(),
                )
            )

            return uri_package_or_wrapper
        except WrapError as err:
            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=uri,
                    description=(
                        f"{self.get_step_description()} - Error: "
                        f"Failed to resolve uri: {uri}, using extension resolver: "
                        f"({self.extension_wrapper_uri})"
                    ),
                    sub_history=sub_context.get_history(),
                )
            )
            raise UriResolverExtensionError(
                f"Failed to resolve uri: {uri}, using extension resolver: "
                f"({self.extension_wrapper_uri})"
            ) from err
        except InfiniteLoopError as err:
            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=uri,
                    description=(
                        f"{self.get_step_description()} - Error: "
                        f"Infinite loop detected when resolving uri: {uri}, "
                        f"using extension resolver: ({self.extension_wrapper_uri})"
                    ),
                    sub_history=sub_context.get_history(),
                )
            )

            if err.uri == self.extension_wrapper_uri:
                raise UriResolverExtensionNotFoundError(
                    self.extension_wrapper_uri
                ) from err
            raise err

    def _try_resolve_uri_with_extension(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a uri or a manifest using the extension wrapper.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            MaybeUriOrManifest: The resolved URI or manifest.
        """
        uri_or_manifest: Optional[MaybeUriOrManifest] = client.invoke(
            uri=self.extension_wrapper_uri,
            method="tryResolveUri",
            args={
                "authority": uri.authority,
                "path": uri.path,
            },
            encode_result=False,
            resolution_context=resolution_context,
        )

        if uri_or_manifest is None:
            return uri

        if result_uri := uri_or_manifest.get("uri"):
            return Uri.from_str(result_uri)

        if result_manifest := uri_or_manifest.get("manifest"):
            package = WasmPackage(
                UriResolverExtensionFileReader(self.extension_wrapper_uri, uri, client),
                result_manifest,
            )
            return UriPackage(uri=uri, package=package)

        return uri


__all__ = ["ExtensionWrapperUriResolver", "MaybeUriOrManifest"]



"""This module contains the resolver for wrappers."""
from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriWrapper,
    Wrapper,
)

from ..abc import ResolverWithHistory


class WrapperResolver(ResolverWithHistory):
    """Defines the wrapper resolver.

    Args:
        uri (Uri): The uri to resolve.
        wrapper (Wrapper): The wrapper to use.
    """

    __slots__ = ("uri", "wrapper")

    uri: Uri
    wrapper: Wrapper

    def __init__(self, uri: Uri, wrapper: Wrapper):
        """Initialize a new WrapperResolver instance."""
        self.uri = uri
        self.wrapper = wrapper
        super().__init__()

    def get_step_description(self) -> str:
        """Get the description of the resolver step.

        Returns:
            str: The description of the resolver step.
        """
        return f"Wrapper ({self.uri})"

    def _try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            UriPackageOrWrapper: The resolved URI, wrap package, or wrapper.
        """
        return uri if uri != self.uri else UriWrapper(uri=uri, wrapper=self.wrapper)


__all__ = ["WrapperResolver"]



"""This module contains all the errors related to URI resolution."""
from polywrap_core import Uri


class UriResolutionError(Exception):
    """Base class for all errors related to URI resolution."""


class InfiniteLoopError(UriResolutionError):
    """Raised when an infinite loop is detected while resolving a URI.

    Args:
        uri (Uri): The URI that caused the infinite loop.
    """

    uri: Uri

    def __init__(self, uri: Uri):
        """Initialize a new InfiniteLoopError instance."""
        self.uri = uri
        super().__init__(
            f"An infinite loop was detected while resolving the URI: {uri.uri}\n"
        )


class UriResolverExtensionError(UriResolutionError):
    """Base class for all errors related to URI resolver extensions."""


class UriResolverExtensionNotFoundError(UriResolverExtensionError):
    """Raised when an extension resolver wrapper could not be found for a URI.

    Args:
        uri (Uri): The URI that caused the error.
    """

    uri: Uri

    def __init__(self, uri: Uri):
        """Initialize a new UriResolverExtensionNotFoundError instance."""
        self.uri = uri
        super().__init__(
            f"Could not find an extension resolver wrapper for the URI: {uri.uri}\n"
        )


__all__ = [
    "UriResolutionError",
    "InfiniteLoopError",
    "UriResolverExtensionError",
    "UriResolverExtensionNotFoundError",
]



"""This module contains the ResolutionResultCacheResolver."""

from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
)

from ...errors import UriResolutionError
from ...types import ResolutionResultCache


class ResolutionResultCacheResolver(UriResolver):
    """An implementation of IUriResolver that caches the URI resolution result.

    The URI resolution result can be a URI, IWrapPackage, Wrapper, or Error.
    Errors are not cached by default and can be cached by setting the cache_errors option to True.

    Args:
        resolver_to_cache (UriResolver): The URI resolver to cache.
        cache (ResolutionResultCache): The resolution result cache.
        cache_errors (bool): Whether to cache errors.
    """

    __slots__ = ("resolver_to_cache", "cache", "cache_errors")

    resolver_to_cache: UriResolver
    """The URI resolver to cache."""

    cache: ResolutionResultCache
    """The resolution result cache."""

    cache_errors: bool
    """Whether to cache errors."""

    def __init__(
        self,
        resolver_to_cache: UriResolver,
        cache: ResolutionResultCache,
        cache_errors: bool = False,
    ):
        """Initialize a new ResolutionResultCacheResolver instance."""
        self.resolver_to_cache = resolver_to_cache
        self.cache = cache
        self.cache_errors = cache_errors

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        This method tries to resolve the URI with the resolver to cache.\
        If the result is in the cache, it returns the cached result.\
        If the result is not in the cache, it resolves the URI using\
        the inner resolver and caches the result.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use.
            resolution_context (UriResolutionContext): The resolution context to use.

        Returns:
            UriPackageOrWrapper: The result of the resolution.
        """
        if cached_result := self.cache.get(uri):
            if isinstance(cached_result, UriResolutionError):
                raise cached_result

            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=cached_result,
                    description="ResolutionResultCacheResolver (Cache)",
                )
            )
            return cached_result

        sub_context = resolution_context.create_sub_history_context()
        result: UriPackageOrWrapper

        try:
            result = self.resolver_to_cache.try_resolve_uri(
                uri,
                client,
                sub_context,
            )
            self.cache.set(uri, result)
        except UriResolutionError as error:
            if self.cache_errors:
                self.cache.set(uri, error)
            resolution_context.track_step(
                UriResolutionStep(
                    source_uri=uri,
                    result=uri,
                    sub_history=sub_context.get_history(),
                    description="ResolutionResultCacheResolver - Error",
                )
            )
            raise error

        resolution_context.track_step(
            UriResolutionStep(
                source_uri=uri,
                result=result,
                sub_history=sub_context.get_history(),
                description="ResolutionResultCacheResolver",
            )
        )
        return result


__all__ = ["ResolutionResultCacheResolver"]



"""This module contains the wrapper cache interface."""
# pylint: disable=unnecessary-ellipsis
from typing import Protocol, Union

from polywrap_core import Uri, UriPackageOrWrapper

from ....errors import UriResolutionError


class ResolutionResultCache(Protocol):
    """Defines a cache protocol for caching resolution results by uri.

    This is used by the resolution result resolver to cache resolution results\
        for a given uri.
    """

    def get(self, uri: Uri) -> Union[UriPackageOrWrapper, UriResolutionError, None]:
        """Get the resolution result from the cache by its uri."""
        ...

    def set(
        self, uri: Uri, result: Union[UriPackageOrWrapper, UriResolutionError]
    ) -> None:
        """Set the resolution result in the cache by its uri."""
        ...

    def __str__(self) -> str:
        """Display cache as a string."""
        ...


__all__ = ["ResolutionResultCache"]



"""This module contains the StaticResolver class."""
from polywrap_core import (
    InvokerClient,
    Uri,
    UriPackage,
    UriPackageOrWrapper,
    UriResolutionContext,
    UriResolutionStep,
    UriResolver,
    UriWrapper,
)

from ...types import StaticResolverLike


class StaticResolver(UriResolver):
    """Defines the static URI resolver.

    Args:
        uri_map (StaticResolverLike): The URI map to use.
    """

    __slots__ = ("uri_map",)

    uri_map: StaticResolverLike

    def __init__(self, uri_map: StaticResolverLike):
        """Initialize a new StaticResolver instance."""
        self.uri_map = uri_map

    def try_resolve_uri(
        self,
        uri: Uri,
        client: InvokerClient,
        resolution_context: UriResolutionContext,
    ) -> UriPackageOrWrapper:
        """Try to resolve a URI to a wrap package, a wrapper, or a URI.

        Args:
            uri (Uri): The URI to resolve.
            client (InvokerClient): The client to use for resolving the URI.
            resolution_context (UriResolutionContext): The resolution context.

        Returns:
            UriPackageOrWrapper: The resolved URI.
        """
        result = self.uri_map.get(uri)

        match result:
            case None:
                description: str = "Static - Miss"
                uri_package_or_wrapper: UriPackageOrWrapper = uri
            case UriPackage():
                description = "Static - Package"
                uri_package_or_wrapper = result
            case UriWrapper():
                description = "Static - Wrapper"
                uri_package_or_wrapper = result
            case _:
                description = f"Static - Redirect ({uri} - {result})"
                uri_package_or_wrapper = result

        step = UriResolutionStep(
            source_uri=uri, result=uri_package_or_wrapper, description=description
        )
        resolution_context.track_step(step)
        return uri_package_or_wrapper


__all__ = ["StaticResolver"]
